static inline __u16 get_reqseq(struct sk_buff *skb)
{
	l2cap_iframe_hdr_t	*iframe = (l2cap_iframe_hdr_t *)(skb->data);
	return (__btoh16(iframe->control) & L2CAP_REQSEQ_MASK)>>8; 
}

static inline void set_reqseq(struct sk_buff *skb, __u16 num)
{
	l2cap_iframe_hdr_t	*iframe = (l2cap_iframe_hdr_t *)(skb->data);
	iframe->control = __htob16(((num<<8) & L2CAP_REQSEQ_MASK) | (__btoh16(iframe->control) & L2CAP_CLEAR_REQSEQ));
}

static inline __u16 get_txseq(struct sk_buff *skb)
{
	l2cap_iframe_hdr_t	*iframe = (l2cap_iframe_hdr_t *)(skb->data);
	return (__btoh16(iframe->control) & L2CAP_TXSEQ_MASK)>>1;
}

static inline void set_txseq(struct sk_buff *skb, __u16 num)
{
	l2cap_iframe_hdr_t	*iframe = (l2cap_iframe_hdr_t *)(skb->data);
	iframe->control = __htob16(((num<<1) & L2CAP_TXSEQ_MASK) | (__btoh16(iframe->control) & L2CAP_CLEAR_TXSEQ));
}

static inline void set_retransmission_bit(struct sk_buff *skb)
{
	l2cap_iframe_hdr_t	*iframe = (l2cap_iframe_hdr_t *)(skb->data);
	iframe->control = __htob16( __btoh16(iframe->control) & L2CAP_SET_RET_DISABLE_BIT);
}

static inline void unset_retransmission_bit(struct sk_buff *skb)
{
	l2cap_iframe_hdr_t	*iframe = (l2cap_iframe_hdr_t *)(skb->data);
	iframe->control = __htob16( __btoh16(iframe->control) & L2CAP_UNSET_RET_DISABLE_BIT);
}

static inline int TxWindowFull(l2cap_ch *ch)
{
	int expected_ack_seq = get_reqseq(ch->expected_ack_seq);
	return (ch->txwindow_size <= (ch->next_tx_seq - expected_ack_seq));
}

void l2cap_onedottwo_init(l2cap_ch *ch)
{
	skb_queue_head_init(&(ch->tx_buffer));
	ch->expected_ack_seq = NULL; 	/* Points to the sk_buff (L2CAP packet) expected to be acknowledge. */
	ch->next_tx_frame = NULL;	/* Points to the sk_buff(L2CAP packet) to be transmited next. */
	ch->next_tx_seq = 0;	
	/* Receiving buffer */
	skb_queue_head_init(&(ch->rx_buffer));
	ch->buffer_seq = NULL;		/* Points to the first sk_buff (L2CAP packet) waiting to be pull out from the upper layer */
	ch->expected_tx_seq = 0;		/* Sequence number of the next packet to be received */ 

	ch->mode = L2CAP_BASIC_MODE;	/* L2CAP mode: 00:Basic, 01: Retransmission, 02: Flow control, other values: RESERVED. */
	ch->txwindow_size = 0; 		/* TxWindow size. Size of the transmission window. */
	ch->max_tx = 0;			/* MaxTransmit. Number of transmission of a single I-frame that L2CAP is allowed. */
	ch->retx_timeout = 0;		/* Retransmission time-out Value in miliseconds of the retransmission time-out. */
	ch->monitor_timeout = 0;	/* Monitor time-out. Value in miliseconds of the interval at which S-frames should be transmitted.*/
	ch->mps = sysctl_l2cap_mtu;	/* Maximum PDU payload size */

	//ch->flow_control = L2CAP_FLOW_CONTROL_OFF;	/* We set initialy the FLOW CONTROL flag to OFF */
	ch->ret_fc_flags = 0;
	/* Reassembly */
	ch->lframe = NULL;		/* Points to the packet that it is currently being reassembled. */

}

__u16 calculate_fcs(struct sk_buff *frame)
{
	l2cap_hdr_t* 	l2cap_frame = (l2cap_hdr_t *)frame->data;
	__u8*  cpos = l2cap_frame->data;
	__u16	len = l2cap_frame->length + 2,gd = 0x8005,lfsr = 0x0000, bit =0x0000, fcs = 0x0000, apply_gd, ind;
	int j;
	
	
	/*NOTE: len has the total size of frame including length and CID fields minus the FCS field */
	
	for(ind=0; ind<= len; ind++){ /* For every byte in the frame (except FCS) do... */
		for(j=0;j<8;j++) { /* From the LSB first apply the algorithm explained in Page 39 of the L2CAP 1.2 Specification */
			bit = (*cpos>>j) & 0x0001;
			apply_gd = ((bit ^ (lfsr>>15)) == 0x0000);
			lfsr = lfsr << 1;
			if (apply_gd) {
				lfsr = lfsr ^ gd;
			}
		}
		cpos+=ind;
	}

	/* The last bit has been entered and now we set the switch to postion 2 (See Figure 3.4, page 39 L2CAP Specification 1.2) to get the FCS value */
	
	for(j=0; j<16; j++)
	{
		fcs = ((lfsr<<j) & 0x8000) | fcs;
	}
	
	return fcs;	
}

int append_unsegmented_hdr(l2cap_ch *ch,struct sk_buff *skb)
{
	unsigned char		*tail = NULL;
	l2cap_iframe_hdr_t	*hdr =  (l2cap_iframe_hdr_t*)skb_push(skb, L2CAP_IFRAME_HDR_LEN);
	int			err = 0;
	
	if ((skb_headroom(skb) >= 6) && (skb_tailroom(skb) >= 2)) {
		hdr->length = __htob16(skb->len+4);	// SDU Len + CONTROL field len + FCS field len
		hdr->cid = __htob16(ch->rcid);
		hdr->control = __htob16(L2CAP_SAR_UNSEGMENTED);		// The rest of the control field is set when sending the frame 
		tail = skb_put(skb,2); 	/* Make space for FCS */
		*(__u16 *)tail = calculate_fcs(skb);
	}
	else
		err = -1;
	
	return err;
}

struct sk_buff *create_start_segmented_iframe(l2cap_ch *ch,struct sk_buff *sdu)
{
	struct sk_buff 		*lskb = NULL;
	l2cap_iframe_hdr_t	*iframe = NULL;
	unsigned char		*ldata = NULL;
	
	lskb = alloc_skb(ch->mps,GFP_ATOMIC);	
	if (lskb == NULL) 
		return NULL;
	
	ldata = skb_put(lskb,8);
	iframe = (l2cap_iframe_hdr_t *)(ldata);
	iframe->length = __htob16(ch->mps - 4);	/* Magic number 4 comes from excluding the length and cid fields from the total length. */
	iframe->cid = __htob16(ch->rcid);
	
	iframe->control = __htob16(L2CAP_SAR_START);	
	*((__u16 *)iframe->data) = __htob16(sdu->len);

	ldata = skb_put(lskb,ch->mps - 10);
	memcpy(ldata, sdu->data, ch->mps - 10);
	
	ldata = skb_put(lskb,2);
	*((__u16 *)ldata) = __htob16(calculate_fcs(lskb));
	
	return lskb;
}

struct sk_buff *create_segmented_iframe(l2cap_ch *ch, __u16 control,unsigned char *sdu, __u16 sdu_len)
{
	struct sk_buff 	*lskb = NULL;
	unsigned char	*ldata = NULL;
	l2cap_iframe_hdr_t *iframe;
	
	lskb = alloc_skb(sdu_len + 8,GFP_ATOMIC);
	if (lskb == NULL)
		return NULL;
	
	ldata =skb_put(lskb,6);
	iframe = (l2cap_iframe_hdr_t *)(ldata);
	iframe->length = __htob16(sdu_len + 8);
	iframe->cid = __htob16(ch->rcid);
	
	iframe->control = __htob16(control);

	ldata = skb_put(lskb,sdu_len);
	memcpy(ldata, sdu, sdu_len);

	ldata = skb_put(lskb,2);
	*(__u16 *)ldata = __htob16(calculate_fcs(lskb));

	return lskb;
}


struct sk_buff *l2cap_reassembly_sdu(l2cap_ch *ch, struct sk_buff *skb)
/* NOTE: This function does not check for mismatches in the frames sequence number, we suppose that before reaching to the reassembly stage the frames are consequitive. */
/* We check the proper frame sequence number upon reception of the packet ! */	
{
	__u16		sdu_len;
	l2cap_iframe_hdr_t *iframe = (l2cap_iframe_hdr_t *)(skb->data);
	
	DBFENTER;

	switch (SAR(iframe->control))
	{
		case L2CAP_SAR_START:
			sdu_len = __btoh16(__get_u16(iframe->data));
			DBPRT("Start L2CAP SDU frame.SDU length: %d\n", sdu_len);
			if (ch->lframe != NULL) {
				DBPRT("Current lframe has not been yet pulled by the upper layers !\n");
				return NULL;	
			}
			ch->lframe = alloc_skb(sdu_len,GFP_ATOMIC);
			if (ch->lframe == NULL) return NULL;
			memcpy(skb_put(ch->lframe,(iframe->length - 10)),(iframe->data+2),(iframe->length - 10));
			break;
		case L2CAP_SAR_CONTINUATION:
		case L2CAP_SAR_END:
			DBPRT("Continuation/End L2CAP SDU frame\n");
			if (skb_tailroom(ch->lframe) < (iframe->length - 8)) {
				DBPRT("L2CAP iFrame too big !. Dropping the packet (=> freeing L2CAP lframe)\n");
				kfree_skb(ch->lframe);
				ch->lframe = NULL;
				return NULL;; 
			}
			memcpy(skb_put(ch->lframe, (iframe->length - 8)), iframe->data, (iframe->length - 8));	
			break;
		default:
			DBPRT("ERROR Unexpected control frame: %x\n",iframe->control);
			break;
	}

	if (SAR(iframe->control) == L2CAP_SAR_END) {
		if (ch->lframe->len == ch->lframe->data_len)
			return ch->lframe;
		else{
			DBPRT("L2CAP lframe size mismatch. Dropping packet (==>freeing L2CAP lframe)\n");
			kfree_skb(ch->lframe);
			ch->lframe = NULL;
		}
	}
	
	DBFEXIT;
	return NULL; 
	//return ((SAR(iframe->control) != L2CAP_SAR_END)? NULL : ch->lframe);
}

struct sk_buff * l2cap_segmentation_sdu(l2cap_ch *ch, struct sk_buff *sdu)
/* This function segmentes the SDU and queue the the segmented L2CAP packets in the transmission queue (tx_buffer) */
{
	struct sk_buff *lskb = NULL;
	struct sk_buff_head llist;
	__u16 offset = 0;
	
	DBFENTER;

	skb_queue_head_init(&llist);

	/* Create segmented START PDU */
	lskb = create_start_segmented_iframe(ch,sdu);
	if (lskb == NULL)
		goto exit;
	skb_queue_tail(&llist, lskb);

	/* Create segmented CONTINUATION PDU */
	offset = lskb->len - 10;
	while (ch->mps < sdu->len -offset) {
		lskb = create_segmented_iframe( ch, L2CAP_SAR_CONTINUATION ,(unsigned char *)(sdu->data + offset), ch->mps);
		if (lskb == NULL)
			goto exit;
		skb_queue_tail(&llist, lskb);
		offset += ch->mps;
	}

	/* Create segmented END PDU */
	lskb = create_segmented_iframe( ch, L2CAP_SAR_END, (unsigned char *)(sdu->data + offset), (sdu->len - offset));
	if(lskb){
		skb_queue_tail(&llist, lskb);
		/* Add L2CAP packets to the transmission queue (tx_buffer) */
		while (!skb_queue_empty(&llist)){
			lskb = skb_peek(&llist);
			skb_unlink(lskb);
			skb_queue_tail(&ch->tx_buffer,lskb);
		}
	}

exit:
	skb_queue_purge(&llist);	
	DBFEXIT;
	return lskb;
}

struct sk_buff* decapsulation(struct sk_buff *skb, __u16 *length, __u16 *cid, __u16 *control, __u16 *sdu_length, __u16* fcs)
{
	
	l2cap_iframe_hdr_t	*hdr = (l2cap_iframe_hdr_t *)skb->data;

	*length = __btoh16(hdr->length);
	*cid = __btoh16(hdr->cid);

	if (skb->len < *length + 4 ) { /* wrong packet */
		BTDEBUG("L2CAP packet too small, drop it...\n");
		kfree_skb(skb);
		return NULL;
	}
	
	if (MODE(ch) == L2CAP_BASIC_MODE){ 
		skb_pull(skb, L2CAP_HDR_LEN);
	}
	else {
		*control = __btoh16(hdr->control)
		*fcs = __btoh16(__get_u16(hdr->tail - 2));
		if (SAR(*control) == L2CAP_SAR_START) {
			*sdu_length = __btoh16(__get_u16(hdr->data));
			skb_pull(skb,2); /* Pull SDU length field */
		}
		skb_pull(skb,2); /* Pull control field */
		skb_trim(skb,skb->len - 2); /* Trim FCS field */
	}

	return skb;
}
